require 'yaml'
require 'fileutils'
require 'shell'

class ConfigurationManager
  include FileUtils
  include MyShell

  def initialize(svn_url, svn_user, svn_password)
    @svn_url = svn_url
    @svn_user = svn_user
    @svn_password = svn_password
    @path = File.join(Dir.getwd, '.sisyphus')
  end

  def synchronize_checkout(time)
    time = svn_time(time)
    remove_existing_checkout
    create_new_checkout(time)
  end

  def create_new_checkout(time)
    if @svn_user == nil and @svn_password == nil then
      svn_command("co #{@svn_url} -r {#{time}} #{@path}")
    elsif @svn_user != nil and @svn_password != nil then
      svn_command("co #{@svn_url} --non-interactive --username #{@svn_user} --password #{@svn_password} -r {#{time}} #{@path}")
    else
      raise RuntimeError.new("User or password given without the other.")
    end
  end

  def remove_existing_checkout
    if test(?d, @path) then
      $stderr << "Warning: #{@path} exists and is removed!\n"
      rm_r @path
    end
  end
  
  def svn_time(time)
    return time.strftime("%Y%m%dT%H%M")
  end
  
  def svn_command(options)
    command = "svn #{options}"
    $stderr << "#{command}\n"
    execute(command)
    #output = `#{command}`
    #if $?.exitstatus != 0 then
    #  raise RuntimeError.new("Subversion exited unsuccessfully: #{output}")
    #end
  end
  
  def version_of(path)
    # assert synchronized here???
    info = read("svn info #{path}")
    #info = `svn info #{path}`
    info.split("\n").each do |line|
      if line =~ /Last Changed Rev: ([0-9]+)/ then
        return $1.to_i
      end
    end
    raise RuntimeError.new("subversion didn't behave as expected: #{info.split.inspect}")
  end

  def rel_to_checkout(path)
    return File.join(@path, path)
  end

  def configuration(profile_name, time)
    synchronize_checkout(time)
    config_dir = rel_to_checkout('config')
    sources = config_item(config_dir, 'sources')
    script = config_item(config_dir, 'script')
    validate_script(script)
    profile = config_item(rel_to_checkout('profiles'), profile_name)
    return Configuration.new(profile, sources, script)
  end

  def validate_script(script)
    missing_entries = []
    script.properties['order'].each do |action|
      if not script.properties['templates'].has_key?(action) then
        missing_entries << action
      end
    end
    if missing_entries != [] then
      raise RuntimeError.new("missing templates for actions: #{missing_entries.inspect}")
    end
    templates_but_no_order = script.properties['templates'].keys - script.properties['order']
    if templates_but_no_order != [] then
      raise RuntimeError.new("templates defined but no actions in order: #{templates_but_no_order}")
    end
  end

  def config_item(dir, name)
    path = File.join(dir, "#{name}.yml")
    props = nil
    File.open(path) do |f|
      props = YAML.load(f)
    end
    version = version_of(path)
    return ConfigurationItem.new(name, version, props)
  end

end


class Configuration
  attr_reader :profile, :sources, :script
  
  def initialize(profile, sources, script)
    @profile = profile
    @sources = sources
    @script = script
  end

  def profile_value(name)
    value = @profile.properties[name]
    if value.nil? then
      raise RuntimeError.new("no #{name} configured in profile: #{@profile}")
    end
    return value    
  end

  def actions
    return script.properties['order'].collect do |action|
      script.properties['templates'][action]
    end
  end

  def eql?(configuration)
    return @profile.eql?(configuration.profile) &&
      @sources.eql?(configuration.sources) &&
      @script.eql?(configuration.script)
  end

  def ==(configuration) 
    return (configuration.profile == @profile) &&
      (configuration.sources == @sources) &&
      (configuration.script == @script)
  end

  def hash
    s = "#{@profile.hash}#{@sources.hash}#{@script.hash}"
    return s.hash
  end

  def to_s
    return "<#{@profile}, #{@sources}, #{@script}>"
  end

  def build_dir
    return profile_value('build_dir')
  end

  def install_dir
    return profile_value('install_dir')
  end

  def user
    return profile_value('user')
  end

end

class ConfigurationItem
  attr_reader :version, :properties, :name

  def initialize(name, version, properties)
    @name = name
    @version = version
    @properties = properties
  end

  def ==(conf_item)
    return (conf_item.name == @name) &&
      (conf_item.version == @version)
  end

  def eql?(conf_item)
    return @name.eql?(conf_item.name) &&
      @version.eql?(conf_item.version)
  end

  def hash
    s = "#{@name.hash}#{@version.hash}"
    return s.hash
  end

  def to_s
    return "<#{@name}, #{@version}>"
  end

end

