
require 'active_record'

class SiComponent < ActiveRecord::Base
  # fields: name

  def prefix(root)
    return File.join(root, name)
  end
end

class SiRevision < ActiveRecord::Base
  belongs_to :si_component
  # fields: version

  def name
    si_component.name
  end

  def prefix(root)
    subdir = version.to_s
    return File.join(si_component.prefix(root), subdir)
  end

  def to_s
    return "#{name}:#{version}"
  end
end


class SiProfile < ActiveRecord::Base
  # fields: name, version
end

class SiHost < ActiveRecord::Base
  # fields: uname
  validates_presence_of :uname

  def name
    if uname =~ /([a-z0-9]+\.[a-z0-9]+\.[a-z0-9]+\.[a-z0-9]+) / then
      return $1
    end
    raise RuntimeError.new("no hostname found in uname: #{uname}")
  end

end

class SiConfig < ActiveRecord::Base
  # fields: sources_version, script_version
  belongs_to :si_profile
end

class SiSession < ActiveRecord::Base
  # fields: time
  has_many :si_items  
end

class SiItem < ActiveRecord::Base
  # fields: success
  belongs_to :si_revision

  ###
  belongs_to :si_host
  belongs_to :si_config # can host/config vary per session? don't think so.
  # move this to session
  # looking for an existing attempt we have one extra indirection:
  # find an item with a session that has the same host and config.

  belongs_to :si_session
  has_many :si_results
  has_and_belongs_to_many :si_deps, 
  :class_name => 'SiItem',
  :join_table => 'si_item_deps',
  :foreign_key => 'si_item_id',
  :association_foreign_key => 'si_dep_id'   


  def add_action_result(command, result)
    r = SiResult.new(:command => command, :output => result.output, :success => result.success)
    si_results << r
  end

  def set_result(result)
    self.success = result.success
  end

  def set_progress(bool)
    self.progress = bool
  end

  def set_success(bool)
    # For some obscure reason, this:
    self.success = bool
    # is not the same as:
    # success = bool (????)
  end

  def in_progress?
    self.progress == true
  end

  def dep_items
    return si_deps
  end

  def name
    return si_revision.name
  end

  def prefix(root)
    subdir = si_session.time.to_i.to_s
    return File.join(si_revision.prefix(root), subdir)
  end
  
  def to_s
    return "<#{si_revision}:#{success}>"
  end

end

class SiResult < ActiveRecord::Base
  # fields: command, output, success
  belongs_to :si_item
end



class DBStore
  def initialize(dbconf)
    @dbconf = dbconf
  end


  def connect
    ActiveRecord::Base.establish_connection(@dbconf)
  end

  def disconnect
    ActiveRecord::Base.remove_connection
  end
  

  def has_item_for_target?(target)
    return contains_revision?(target.host, target.config, target.revision)
  end

  def last_item_for_target(target)
    return latest_item(target.host, target.config, target.revision)
  end

  def item_for_target(target, dep_items)
    h, c, r = host_config_revision(target.host, target.config, target.revision)
    s = db_session(target.time)
    item = SiItem.new(:si_config => c, :si_host => h, :si_revision => r, 
                      :si_session => s, :si_deps => dep_items,
                      :success => false, :progress => true)
    item.save
    return item
  end


  ## Refactor the three below!!!
  def latest_item(host, config, revision)
    h, c, r = host_config_revision(host, config, revision)
    item = SiItem.find(:first, 
                       :conditions => ['si_host_id = ? and si_config_id = ? and si_revision_id = ?',
                         h.id, c.id, r.id],
                       :order => 'id asc')
    if item.nil? then
      raise RuntimeError.new("latest_item called but revision is not available: #{revision}")
    end
    return item
  end


  def contains_revision?(host, config, revision)
    h, c, r = host_config_revision(host, config, revision)
    item = SiItem.find(:first, 
                       :conditions => ['si_host_id = ? and si_config_id = ? and si_revision_id = ?',
                         h.id, c.id, r.id])
    if item.nil? then
      return false      
    end
    return true
  end

  def make_item(host, config, revision, dep_items, success = false)
    h, c, r = host_config_revision(host, config, revision)
    item = SiItem.new(:si_config => c, :si_host => h, :si_revision => r, 
                      :si_session => @session, :si_deps => dep_items,
                      :success => success)
    return item
  end


  def save(item)
    SiItem.transaction do 
      item.save
    end
  end

  def time_of_last_session
    session = SiSession.find(:first, :order => 'time asc')
    if session.nil? then
      return nil
    end
    return session.time
  end

  def host_config_revision(host, config, revision)
    h = db_host(host)
    c = db_config(config)
    r = db_revision(revision)
    return h, c, r
  end

  def db_host(host)
    h = SiHost.find(:first, :conditions => ['uname = ?', host])
    if h.nil? then
      h = SiHost.new(:uname => host)
    end
    return h
  end

  def db_session(time)
    s = SiSession.find(:first, :conditions => ['time = ?', time])
    if s.nil? then
      s = SiSession.new(:time => time)
    end
    return s
  end

  def db_component(component)
    name = component.name
    c = SiComponent.find(:first, :conditions => ['name = ?', name])
    if c.nil? then
      c = SiComponent.new(:name => name)
    end
    return c
  end

  def db_revision(revision)
    c = db_component(revision.component)
    r = SiRevision.find(:first, :conditions => ['si_component_id = ? and version = ?', 
                          c.id, revision.version])
    if r.nil? then
      r = SiRevision.new(:si_component => c, :version => revision.version)
    end
    return r
  end

  def db_profile(profile)
    p = SiProfile.find(:first, :conditions => ['name = ? and version = ?',
                         profile.name, profile.version])
    if p.nil? then
      p = SiProfile.new(:name => profile.name, :version => profile.version)
    end
    return p
  end

  def db_config(config)
    p = db_profile(config.profile)
    sov = config.sources.version
    scv = config.script.version
    c = SiConfig.find(:first, :conditions => 
                        ['si_profile_id = ? and sources_version = ? and script_version = ?',
                      p.id, sov, scv])
    if c.nil? then
      c = SiConfig.new(:si_profile => p, :sources_version => sov,:script_version => scv)
    end
    return c
  end    

end



if __FILE__ == $0 then
  conf = {
    'adapter' => 'postgresql',
    'database' => 'sisyphusdev',
    'host' =>  'sjofar.sen.cwi.nl',
    'username' => 'sisyphusdev',
    'password' => 'tantalu5'
  }
  ActiveRecord::Base.establish_connection(conf)
  c.save
end
