require 'fileutils'
require 'tempfile'
require 'myshell'
require 'time'

class RepositoryFactory
  def repository(name, repositories)
    repositories.each do |repo|
      if repo['components'].include?(name) then
        return make_repository(repo['type'], repo['protocol'], repo['location'])
      end
    end
    raise RuntimeError.new("cannot find module #{name} in #{repositories.inspect}")
  end

  def make_repository(type, protocol, location)
    if type == 'cvs' then
      return CVSRepository.new(protocol, location)
    elsif type == 'subversion' then
      return SVNRepository.new(protocol, location)
    end
    raise RuntimeError.new("unknown repository type: #{type}")    
  end
end


class Repository
  def initialize(protocol, location)
    @protocol = protocol
    @location = location
  end
end

class CVSRepository < Repository
  include FileUtils
  include MyShell

  def checkout(user, time, modulename, targetdir, dirname = modulename)
    cvs_time = convert_to_gmt(time)
    path = File.join(targetdir, dirname)
    cd(targetdir) do
      if test(?d, dirname) then
        $stderr << "WARNING: checkout exists: #{path} removed!\n"
        rm_r dirname
      end
      cvs_command(user, "co -D \"#{cvs_time}\" #{modulename}")
      if modulename != dirname then
        mv(modulename, dirname)
      end
    end
    return CVSCheckout.new(self, path)
  end

  def convert_to_gmt(time)
    cvs_time = time.clone
    cvs_time.gmtime
    return cvs_time
  end

  def empty_file?(path)
    return !test(?s, path)
  end

  def cvs_url(user)
    if @protocol == 'ssh' then
      return ":ext:#{user}@#{@location}"
    end
    raise RuntimeError.new("unsupported CVS protocol: #{@protocol}")
  end

  def cvs_command(user, options)
    url = cvs_url(user)
    command = "cvs -Q -d #{url} #{options}"
    puts command
    execute(command)
    # `#{command}`
    #if $?.exitstatus != 0 then
    #  raise RuntimeError.new("execution of '#{command}' failed: #{$!}")
    #end
  end
  
  def to_s
    return "cvs+#{@protocol}://#{@location}"
  end

end

class SVNRepository
  # Todo
end

class Checkout
  attr_reader :path, :repository

  def initialize(repository,  path)
    @repository = repository
    @path = path
  end

  def destroy
    #`rm -r #{@path}`
    $stderr << "WARNING: removal temporarily disabled: rm -r #{@path}\n"
  end

  def version
    raise RuntimeError.new("class Checkout is abstract!")
  end

  def extract_deps
    path = depfile_path(@path)
    File.open(path) do |file|
      return parse_deps(file.read)
    end
  end

  def depfile_path(root)
    path = File.join(root, 'package')
    if not test(?f, path) then
      pattern = File.join(root, "*.pkg")
      pkgs = Dir[pattern]
      path = pkgs == [] ? nil : pkgs[0]
    end
    if path.nil? then
      raise RuntimeError.new("no dependency file found in #{root}!")
    end
    return path
  end 

  def parse_deps(packagedef)
    requires = packagedef
    requires.sub!(/.*requires/m, '')
    requires.sub!(/configuration.*/m, '')
    requires.squeeze!("\n")
    requires.strip!
    return remove_versions(requires.split("\n"))
  end

  def remove_versions(deps)
    return deps.collect do |dep|
      dep.split[0]
    end
  end  

end

class CVSCheckout < Checkout
  def version
    return time_of_last_commit.to_i
  end

  def time_of_last_commit
    sorted_dates = entry_dates(@path).sort
    return sorted_dates.at(-1)
  end

  def entry_dates(path)
    dates = []
    pattern = File.join(path, '**', 'Entries')
    Dir[pattern].each do |entries_path|
      dates += parse_dates(entries_path)
    end
    return dates
  end  
  
  def parse_dates(entries_path)
    dates = []
    File.open(entries_path) do |f|
      f.each_line do |line|
        if  /^\/.*\/.*\/(.*) ([0-9]{4})\/.*\// =~ line then
          dates << Time.parse("#{$1} GMT #{$2}")
        end
      end
    end
    return dates
  end   
end

class SVNCheckout
  include MyShell

  def version
    info = read("svn info #{@path}")
    info.split("\n").each do |line|
      if line =~ /Last Changed Rev: ([0-9]+)/ then
        return $1.to_i
      end
    end
    raise RuntimeError.new("subversion didn't behave as expected: #{info.split.inspect}")
  end
end
